# -*- coding: utf-8 -*-
#
# (c) Sven Sager, License: GPLv3
#
# -*- coding: utf-8 -*-
"""Globaler Funktionskatalog fuer den Revolutionsumbau."""
import revpimodio2
from threading import Thread, Event


class Anlage(Thread):

    """Anlagenthread zur ueberwachung genereller Eigenschaften.

    Dieser Thread lauft dauerhaft mit dem eigentlichen Programm und laesst
    die LED A1 am RevPi-Core blinken. Damit kann ueberwacht werden, ob das
    Programm noch laeuft.

    """

    dict_stoer = {
        "fu_ok": "fu fehler",
        "m_druck_ok": "druckverlust oder nicht ausreichender druck",
        "motorschutz_ok": "motorschutschalter abgefallen",
        "notaus_ok": "notaus ausgelöst",

        "err_endlage": "eine endlage wurde nicht erreicht",
        "err_funichtaus": "fu nicht angehalten",
        "err_magazin1": "magazin 1 nicht erreicht",
        "err_magazin2": "magazin 2 nicht erreicht",
        "err_rutsche": "sensor wird nicht frei",
    }

    def __init__(self, rpi):
        """Instantiiert den Anlagenthread."""
        super().__init__()
        self.evt_ende = Event()
        self.daemon = True
        self.rpi = rpi

    def checkmagazin(self, ioname=None, iovalue=None):
        """Verwaltet Magazinstatus und Warnungen."""
        # Magazin 1
        self.rpi.io.warn_mag1full.value = \
            self.rpi.io.magazin1.value >= self.rpi.io.magazin1_max.value

        # Magazin 2
        self.rpi.io.warn_mag2full.value = \
            self.rpi.io.magazin2.value >= self.rpi.io.magazin2_max.value

    def evt_t_quit(self, ioname, iovalue):
        """Stoerung quittieren Meldung."""
        self.rpi.io.fu_reset.value = iovalue

        # Störmeldungen auf False setzen
        for st in Anlage.dict_stoer:
            if self.rpi.io[st].type == revpimodio2.OUT:
                self.rpi.io[st].value = False

        # Versuchen Masterfehler zu quittieren
        self.rpi.io.err_master.value = not (
            self.rpi.io.notaus_ok.value
            and self.rpi.io.motorschutz_ok.value
            and self.rpi.io.fu_ok.value
            and self.rpi.io.m_druck_ok.value
        )

    def evt_p_mberry(self, ioname, iovalue):
        """Reicht den Fernwarnungsanforderung durch."""
        self.rpi.io.mb_connect.value = iovalue

    def resetmagazin(self, ioname, iovalue):
        """Eventfunktion fuer Reset von Magazin."""
        if ioname == "t_magazin1":
            self.rpi.io.magazin1.value = 0
        if ioname == "t_magazin2":
            self.rpi.io.magazin2.value = 0

    def run(self):
        """Wird durchgehend ausgefuehrt, wenn Anlage lauft."""

        # Event für Störungsquit und Fernwartung
        self.rpi.io.t_quit.reg_event(self.evt_t_quit)
        self.rpi.io.p_mberry.reg_event(self.evt_p_mberry)

        # Events für den Reset der Magazine registrieren
        self.rpi.io.t_magazin1.reg_event(
            self.resetmagazin, edge=revpimodio2.RISING
        )
        self.rpi.io.t_magazin2.reg_event(
            self.resetmagazin, edge=revpimodio2.RISING
        )

        # Events für Füllstandsänderungen registrieren
        self.rpi.io.magazin1.reg_event(self.checkmagazin)
        self.rpi.io.magazin1_max.reg_event(self.checkmagazin)
        self.rpi.io.magazin2.reg_event(self.checkmagazin)
        self.rpi.io.magazin2_max.reg_event(self.checkmagazin)

        # Werte laden
        self.checkmagazin()

        # Versuchen Störungen zu löschen
        self.evt_t_quit("", False)

        while not self.evt_ende.is_set():
            # LED am Core
            self.rpi.core.A1 = int(myflash.status2)
            if self.rpi.io.err_master.value:
                self.rpi.io.h_fehler.value = myflash.status1
            else:
                self.rpi.io.h_fehler.value = False

            myflash.wait()

        # Alle registrierten Events entfernen
        self.rpi.io.t_magazin1.unreg_event()
        self.rpi.io.t_magazin2.unreg_event()
        self.rpi.io.magazin1.unreg_event()
        self.rpi.io.magazin1_max.unreg_event()
        self.rpi.io.magazin2.unreg_event()
        self.rpi.io.magazin2_max.unreg_event()

    def set_stoerung(self, value):
        """Setzt Stoerungsflag, bei False pruefen ob moeglich.
        @param value Stoerungsflag"""
        # Masterfehler generell setzen
        self.rpi.io.err_master.value = True

        # Fehlerbits auswerten
        if value in self.rpi.io \
                and self.rpi.io[value].type == revpimodio2.OUT:
            self.rpi.io[value].value = True

        # Fehlermeldung ausgeben
        if value in Anlage.dict_stoer:
            print("Fehler '{}' ausgelöst".format(Anlage.dict_stoer[value]))
        else:
            print("Unbekannter Fehler durch {}".format(value))

    def stop(self):
        """Beendet den Anlagenthread."""
        self.evt_ende.set()


class Band(Thread):

    """Thread fuer die Steuerung des Bandmotors.

    Ueber die hier zur Verfuegung gestellten Funktionen steuert das Programm
    die Eigenschaften des Bandmotors. Die Funktionen kontrollieren
    selbststaendig ob die gewuenschte Betriebsart erlaubt ist und geben jeweils
    True oder False zurueck. Damit kann in dem aufrufenden Teiles des
    Programms kontrolliert werden, ob die Aktion durchgefuert werden konnte.

    """

    def __init__(self, rpi):
        """Instantiiert den Bandthread."""
        super().__init__()
        self.evt_ende = Event()
        self.daemon = True
        self.rpi = rpi

    def grundstellung(self, wait=False):
        """Setzt die Motorensteuerung in die Grundstellung.
        @return True, wenn erfolgreich durchgefuert"""
        self.rpi.io.fu_rechts.value = False
        self.rpi.io.fu_links.value = False
        self.rpi.io.fu_schnell.value = False
        self.rpi.io.fu_lahm.value = False
        self.rpi.io.fu_soll.value = 0
        if wait:
            return self.rpi.io.fu_motoran.wait(
                exitevent=self.evt_ende,
                okvalue=False,
                timeout=2000
            ) <= 0
        else:
            return True

    def set_lahm_rechts(self):
        """Band langsam nach rechts laufen lassen.
        @return True, wenn Aktion ausgefuehrt werden konnte"""
        return self.switch("fu_rechts", 5000)

    def set_lahm_links(self):
        """Band langsam nach links laufen lassen.
        @return True, wenn Aktion ausgefuehrt werden konnte"""
        return self.switch("fu_links", 5000)

    def set_schnell_rechts(self):
        """Band schnell nach rechts laufen lassen.
        @return True, wenn Aktion ausgefuehrt werden konnte"""
        return self.switch("fu_rechts", 10000)

    def set_schnell_links(self):
        """Band schnell nach links laufen lassen.
        @return True, wenn Aktion ausgefuehrt werden konnte"""
        return self.switch("fu_links", 10000)

    def switch(self, richtung, speed):
        """Steuert FU."""
        if self.rpi.io.notaus_ok.value and self.rpi.io.fu_frei.value:

            # Wenn bereits läuft
            if self.rpi.io[richtung].value and \
                    self.rpi.io.fu_soll.value == speed:
                return True

            # Grundstellung einnehmen und bei Erfolg schalten
            if self.grundstellung(self.rpi.io.fu_motoran.value):
                self.rpi.io[richtung].value = True
                self.rpi.io.fu_soll.value = speed
                return True

        else:
            self.grundstellung()
        return False

    def run(self):
        """Wird durchgehend ausgefuehrt, wenn Bandthread laeuft.

        In dieser Funktion werden die Lampen gesteuert, die je nach Betriebsart
        dauerhaft leuchten (Automatikbetrieb) oder blinken (Handbetrieb) wenn
        der Motor lauft.

        """
        while not self.evt_ende.is_set():
            # FU-Freigabe auswerten
            self.rpi.io.fu_frei.value = \
                self.rpi.io.notaus_ok.value and \
                self.rpi.io.motorschutz_ok.value

            myflash.wait()

        # FU-Freigabe entziehen
        self.rpi.io.fu_frei.value = False

    def stop(self):
        """Beendet den Bandthread."""
        self.evt_ende.set()


class Flasher(Thread):

    """Flasher Thread, der fuer ein synchrones Blinken der Lampen sorgt."""

    def __init__(self, minticksek):
        """Instantiiert Flasher-Thread."""
        super().__init__()
        self.daemon = True
        self._exit = Event()
        self._mintick = minticksek
        self._tick = Event()

        # Status für einfache Zeit
        self.status1 = False

        # Status für zweifache Zeit
        self.status2 = False

        # Status für dreifache Zeit
        self.status3 = False

        # Status für vierfache Zeit
        self.status4 = False

        # Thread startet sich selbst
        self.start()

    def run(self):
        """Laeuft im angegebenen Takt durch und setzt Stati."""
        s3c = 0
        while not self._exit.is_set():
            s3c += 1

            if self.status2 and self.status1:
                self.status4 = not self.status4
            if s3c == 3:
                self.status3 = not self.status3
                s3c = 0
            if self.status1:
                self.status2 = not self.status2
            self.status1 = not self.status1
            self._exit.wait(self._mintick)
            self._tick.set()

        # Aufräumen
        self.status1 = False
        self.status2 = False
        self.status3 = False
        self.status4 = False

    def stop(self):
        """Beendet den Flahserthread."""
        self._exit.set()

    def wait(self):
        """ Wartet auf einen Takttick.
        @return True, wenn erfolgreich, False bei Zeitueberschreitung"""
        tk = self._tick.wait(self._mintick * 2)
        self._tick.clear()
        return tk


class Zylinder(Thread):

    """Thread fuer die Steuerung der Zylinder.

    Ueber die hier zur Verfuegung gestellten Funktionen steuert das Programm
    die Eigenschaften der Zylinder.

    """

    def __init__(self, rpi):
        """Instantiiert den Zylinder-Thread."""
        super().__init__()
        self.evt_ende = Event()
        self.daemon = True
        self.rpi = rpi

    def grundstellung(self):
        """Setzt die Zylinder in die Grundstellung."""
        self.set_m1(False)
        self.set_m2(False)
        self.set_m3(False)

    def set_m1(self, raus, wait=False, exitevent=None):
        """Zylinder m1 Ein-/Ausfahren.

        @param raus Bei True ausfahren / Bei False einfahren
        @param wait Wenn True auf Endlage warten
        @param exitevent Event zum Abbrechen von wait
        @return Wait-Wert oder 0

        """
        # Ventil stellen
        if self.rpi.io.notaus_ok.value:
            self.rpi.io.v_m1_einfahren.value = not raus
            self.rpi.io.v_m1_ausfahren.value = raus

        # Auf Endlage warten
        if wait:
            if raus:
                return self.rpi.io.m_m1_ausgefahren.wait(
                    exitevent=exitevent, timeout=2000
                )
            else:
                return self.rpi.io.m_m1_eingefahren.wait(
                    exitevent=exitevent, timeout=2000
                )
        return 0

    def set_m2(self, raus, wait=False, exitevent=None):
        """Zylinder m2 Ein-/Ausfahren.

        @param raus Bei True ausfahren / Bei False einfahren
        @param wait Wenn True auf Endlage warten
        @param exitevent Event zum Abbrechen von wait
        @return Wait-Wert oder 0

        """
        # Ventil stellen
        if self.rpi.io.notaus_ok.value:
            self.rpi.io.v_m2_einfahren.value = not raus
            self.rpi.io.v_m2_ausfahren.value = raus

        # Auf Endlage warten
        if wait:
            if raus:
                return self.rpi.io.m_m2_ausgefahren.wait(
                    exitevent=exitevent, timeout=2000
                )
            else:
                return self.rpi.io.m_m2_eingefahren.wait(
                    exitevent=exitevent, timeout=2000
                )
        return 0

    def set_m3(self, raus, wait=False, exitevent=None):
        """Zylinder m3 Ein-/Ausfahren.

        @param raus Bei True ausfahren / Bei False einfahren
        @param wait Wenn True auf Endlage warten
        @param exitevent Event zum Abbrechen von wait
        @return Wait-Wert oder 0

        """
        # Ventil stellen
        if self.rpi.io.notaus_ok.value:
            self.rpi.io.v_m3_einfahren.value = not raus
            self.rpi.io.v_m3_ausfahren.value = raus

        # Auf Endlage warten
        if wait:
            if raus:
                return self.rpi.io.m_m3_ausgefahren.wait(
                    exitevent=exitevent, timeout=2000
                )
            else:
                return self.rpi.io.m_m3_eingefahren.wait(
                    exitevent=exitevent, timeout=2000
                )
        return 0

    def run(self):
        """Wird durchgehend ausgefuehrt, wenn Zylinderthread lauft.

        In dieser Funktion werden die Lampen gesteuert, die je nach
        Zylinderstellung dauerhaft leuchten (in Grundstellung) oder
        blinken (nicht in Grundstellung).

        """
        while not self.evt_ende.is_set():
            # Auswerungsarbeiten hier erledigen
            myflash.wait()

    def stop(self):
        """Beendet den Zylinderthread."""
        self.evt_ende.set()


# Globale Funktionen registrieren
anlage = None
band = None
zylinder = None

# Globalen Flasher mit 250 Millisekunden Instantiieren
myflash = Flasher(0.25)
