#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# (c) Sven Sager, License: GPLv3
#
"""Programm fuer unseren Revolutionsumbau."""
import funccatalog
import revpimodio2
import signal
from queue import Queue
from threading import Thread, Event


class AutomatikPara(Thread):

    """Thread fuer den Automatikbetrieb."""

    def __init__(self, rpi):
        """Instantiiert den Automatik-Thread."""
        super().__init__()
        self.clean = False
        self.evt_ende = Event()
        self.daemon = True
        self.mrkmag1 = 0
        self.mrkmag2 = 0
        self.mrkrutsch = 0
        self.rpi = rpi
        self.schlange1 = Queue()
        self.schlange2 = Queue()

    def evt_new(self, th):
        """Fuegt einen neuen Wuerfel in die Queue ein.
        @param th EventCallback-Thread"""
        self.mrkrutsch = 0

        # Band starten
        funccatalog.band.set_lahm_rechts()

        if self.rpi.io.s_metall.value:
            print("metall erkannt")
            self.schlange1.put_nowait(False)
        else:
            print("kunststoff erkannt")
            self.schlange1.put_nowait(True)

        # Auf Verlassen der Sensoren warten
        ec = self.rpi.io.s_rutsche.wait(
            exitevent=self.evt_ende,
            edge=revpimodio2.FALLING,
            okvalue=False,
            timeout=4000
        )
        if ec == 2:
            funccatalog.anlage.set_stoerung("err_rutsche")
            self.evt_ende.set()

    def evt_magazin(self, th):
        """Neuen Stein aus Magazin holen.
        @param th EventCallback-Thread
        @return None"""

        # Zylinder m1 ausfahren und auf Endlage warten
        ec = funccatalog.zylinder.set_m1(True, True, self.evt_ende)
        if ec > 0:
            if ec == 2:
                funccatalog.anlage.set_stoerung("err_endlage")
            self.evt_ende.set()
            return None

        # Zylinder m1 einfahren
        funccatalog.zylinder.set_m1(False)

    def evt_sort1(self, th):
        """Bearbeitet Element bei s_magazin1.
        @param th EventCallback-Thread
        @return None"""

        # Merker aus Queue holen
        try:
            mag2 = self.schlange1.get(timeout=0.2)
        except:
            return None
        self.mrkmag1 = 0

        # Wenn Magazin voll ist einfach raus fahren
        # TODO: anders
        if self.checkmagazinvoll(mag2):
            # Durchfahrt bei Magazin 2 anmelden
            self.schlange2.put_nowait(False)

            # Wenn letzter Stein, dann warten
            if self.schlange1.empty():
                ec = self.rpi.io.s_rutsche.wait(
                    edge=revpimodio2.RISING,
                    exitevent=self.evt_ende,
                    timeout=3000
                )
                if ec > 0:
                    # Wenn noch immer letzter Stein, dann Motor aus
                    print("magazinvoll grundstellung")
                    funccatalog.band.grundstellung()
                else:
                    print("magazinvoll keine grundstellung")
            else:
                print("schlange 1 nicht leer")
            return None

        ec = 0
        if mag2:
            # Magazin 2 Objekt an nächste Schlange reichen
            self.schlange2.put_nowait(mag2)
            return None

        # Magazinzähler Kunststoff erhöhen
        self.rpi.io.magazin1.value += 1

        # Zylinder m2 ausfahren und warten auf Endlage
        print("auswerfen")
        ec = funccatalog.zylinder.set_m2(True, True, self.evt_ende)
        if ec > 0:
            if ec == 2:
                funccatalog.anlage.set_stoerung("err_endlage")
            self.evt_ende.set()
            return None

        # Zylinder m2 einfahren
        funccatalog.zylinder.set_m2(False)

        # Wenn keine Items vorhanden, Band stoppen
        if self.schlange1.empty() and self.schlange2.empty():
            print("motor aus")
            funccatalog.band.grundstellung()

    def evt_sort2(self, th):
        """Bearbeitet Element bei s_magazin2.
        @param th EventCallback-Thread
        @return None"""
        # Merker aus Queue holen
        try:
            if not self.schlange2.get(timeout=0.2):
                # Durchfahrt, nicht auswerfen
                return None
        except:
            return None
        self.mrkmag2 = 0

        # Magazinzähler Metall erhöhen
        self.rpi.io.magazin2.value += 1

        # Zylinder m3 ausfahren und warten auf Endlage
        print("auswerfen")
        ec = funccatalog.zylinder.set_m3(True, True, self.evt_ende)
        if ec > 0:
            if ec == 2:
                funccatalog.anlage.set_stoerung("err_endlage")
            self.evt_ende.set()
            return None

        # Zylinder m3 einfahren
        funccatalog.zylinder.set_m3(False)

        # Wenn keine Items vorhanden, Band stoppen
        if self.schlange1.empty() and self.schlange2.empty():
            print("motor aus")
            funccatalog.band.grundstellung()

    def run(self):
        """Wird durchgehend ausgefuehrt, wenn Anlage im Automatikbetrieb.

        Diese Funktion lauft durchgehend und Regelt die Ablaeufe fuer den
        Automatikbetrieb. Die wait() Funktionen bekommen immer das Exit-Event
        self.evt_ende uebergeben, damit diese beim Beenden vom Automatikbetrieb
        abgebrochen werden. Da als Wert >0 zurueckgegeben wird, brechen wir
        die Verarbeitung an der Stelle ab. Die while-Schleife wird dann
        verlassen.

        Ungewoehnlich ist evtl. das sich das Programm DAUERHAFT in dieser
        Schleife aufhaelt. Die automatische Prozessbildaktualisierung von
        RevPiModIO synchronisiert im Hinterhund staendig unsere Inputs und
        Outputs. Und ein Blockieren des Programms verhindern wir damit, dass
        diese Funktion als Thread ausgefuehrt wird.

        """
        # Lampe h_start Einschalten
        self.rpi.io.h_start.value = True
        self.rpi.io.aktivauto.value = True

        # Automatisch Grundstellung herstellen
        funccatalog.band.grundstellung()
        funccatalog.zylinder.grundstellung()

        # Neuer Stein an Sensor Event
        self.rpi.io.s_rutsche.reg_event(
            self.evt_new, edge=revpimodio2.RISING, as_thread=True
        )

        # Verarbeitungsevents
        self.rpi.io.s_magazin1.reg_event(
            self.evt_sort1, edge=revpimodio2.RISING, as_thread=True
        )
        self.rpi.io.s_magazin2.reg_event(
            self.evt_sort2, edge=revpimodio2.RISING, as_thread=True
        )

        # Schon ein Stein vorm Sensor?
        if self.rpi.io.s_rutsche.value:
            self.evt_new(None)

        # Neuen Stein aus Magazin holen
        if self.rpi.io.p_schnell.value:
            self.rpi.io.s_rutsche.reg_event(
                self.evt_magazin, edge=revpimodio2.FALLING, as_thread=True
            )

            # Ersten stein
            self.evt_magazin(None)

        while not self.evt_ende.is_set():

            if not self.schlange1.empty():
                if self.mrkmag1 == 6:
                    self.evt_ende.set()
                    funccatalog.anlage.set_stoerung("err_magazin1")
                self.mrkmag1 += 1

            if not self.schlange2.empty():
                if self.mrkmag2 == 3:
                    self.evt_ende.set()
                    funccatalog.anlage.set_stoerung("err_magazin2")
                self.mrkmag2 += 1

            if self.rpi.io.p_schnell.value \
                    and self.schlange1.empty() \
                    and self.schlange2.empty():
                if self.mrkrutsch == 5:
                    self.evt_ende.set()
                self.mrkrutsch += 1

            self.evt_ende.wait(1)

        # Events löschen
        self.rpi.io.s_rutsche.unreg_event()
        self.rpi.io.s_magazin1.unreg_event()
        self.rpi.io.s_magazin2.unreg_event()

        # Grundstellung einnehmen
        funccatalog.band.grundstellung()
        if self.clean:
            funccatalog.zylinder.grundstellung()

        # Lampen ausschalten
        self.rpi.io.h_start.value = False
        self.rpi.io.aktivauto.value = False

    def checkmagazinvoll(self, mag2):
        """Prueft ob eines der Magazine voll ist.
        @return True, wenn eines der Magazine voll ist"""
        return (
            mag2 and self.rpi.io.warn_mag2full.value or
            not mag2 and self.rpi.io.warn_mag1full.value
        )

    def stop(self):
        """Beendet den Automatikbetrieb."""
        self.clean = True
        self.evt_ende.set()


class Hand(Thread):

    """Thread fuer den Handbetrieb."""

    def __init__(self, rpi):
        """Instantiiert Thread."""
        super().__init__()
        self.evt_ende = Event()
        self.daemon = True
        self.rpi = rpi

        # Merker setzen
        self.m1 = self.rpi.io.m_m1_ausgefahren.value
        self.m2 = self.rpi.io.m_m2_ausgefahren.value
        self.m3 = self.rpi.io.m_m3_ausgefahren.value

    def run(self):
        """Wird ausgefuehrt wenn der Thread gestartet wird.

        Hier werden fuer alle Taster des Handbetriebs die noetigen Events
        registriert. Die einzige Aufgabe, die diese Funktion sonst noch
        ausfuehrt, ist das Blinken der Lampe h_automatik. Alles weitere wird
        durch Events gesteuert.

        """
        # Panel Buttons
        self.rpi.io.t_m1.reg_event(self.evt_zylinder, edge=revpimodio2.RISING)
        self.rpi.io.t_m2.reg_event(self.evt_zylinder, edge=revpimodio2.RISING)
        self.rpi.io.t_m3.reg_event(self.evt_zylinder, edge=revpimodio2.RISING)
        self.rpi.io.t_motorstart.reg_event(self.evt_bandlauf)

        # Warten bis Ende
        self.rpi.io.aktivhand.value = True
        while not self.evt_ende.is_set():
            self.rpi.io.h_start.value = funccatalog.myflash.status3
            funccatalog.myflash.wait()

        # Motor abschalten
        funccatalog.band.grundstellung()

        # Lampen ausschalten
        self.rpi.io.h_start.value = False
        self.rpi.io.aktivhand.value = False

    def evt_bandlauf(self, ioname, iovalue):
        """Eventfunktion fuer das Ein-/Auschalten des Bandmotors."""
        aktion = False
        if iovalue:
            if self.rpi.io.t_rechts.value:
                # Rechtslauf
                aktion = funccatalog.band.switch(
                    "fu_rechts", self.rpi.io.p_drehzahl.value
                )
            else:
                # Linkslauf
                aktion = funccatalog.band.switch(
                    "fu_links", self.rpi.io.p_drehzahl.value
                )

        if not aktion:
            funccatalog.band.grundstellung()

    def evt_zylinder(self, ioname, iovalue):
        """Eventfunktion fuer Ein-/Ausfahren von Zylinder."""
        if ioname == "t_m1":
            self.m1 = not self.m1
            funccatalog.zylinder.set_m1(self.m1)
        elif ioname == "t_m2":
            self.m2 = not self.m2
            funccatalog.zylinder.set_m2(self.m2)
        elif ioname == "t_m3":
            self.m3 = not self.m3
            funccatalog.zylinder.set_m3(self.m3)

    def stop(self):
        """Beendet den Thread und entfernt alle registrierten Events.

        Durch das Entfernen der Events sind die Taster fuer den Handbetrieb
        ohne Funktion.

        """
        self.evt_ende.set()

        # Panel Buttons
        self.rpi.io.t_m1.unreg_event()
        self.rpi.io.t_m2.unreg_event()
        self.rpi.io.t_m3.unreg_event()
        self.rpi.io.t_motorstart.unreg_event()


class RevPiRevolution():

    """Hauptklasse fuer den Revolutionsumbau."""

    def __init__(self):
        """Instantiiert diese Klasse."""
        self.rpi = None

        # Gesamte Logik für den Automatikbetrieb
        self.th_automatik = None

        # Gesamte Logik für den Handbetrieb
        self.th_hand = None

        # Alles Konfigurieren
        self.loadconfig()

        # Signale zum Beenden einrichten
        signal.signal(signal.SIGINT, self.stop)
        signal.signal(signal.SIGTERM, self.stop)

    def loadconfig(self):
        """Konfiguriert alles fuer die Ausfuehrung."""

        # RevPiModIO konfigurieren mit autorefresh
        self.rpi = revpimodio2.RevPiModIO(autorefresh=True, syncoutputs=False)
        self.rpi.cycletime = 20

        # Für virtuelles Device piCtory Input-Defaults laden
        self.rpi.device.virt01.writeinputdefaults()

        # Alle zusätzlichen IOs für den revpi hier zentral anlegen

        # Panel Eingänge Byte 0
        self.rpi.io.pbit0_7.replace_io("p_online", "?", bit=0)
        self.rpi.io.pbit0_7.replace_io("p_mberry", "?", bit=6)
        self.rpi.io.pbit0_7.replace_io("p_schnell", "?", bit=7)

        # Panel Eingänge Byte 1
        self.rpi.io.pbit8_15.replace_io("t_m1", "?", bit=0)
        self.rpi.io.pbit8_15.replace_io("t_m2", "?", bit=1)
        self.rpi.io.pbit8_15.replace_io("t_m3", "?", bit=2)
        self.rpi.io.pbit8_15.replace_io("t_rechts", "?", bit=3)
        self.rpi.io.pbit8_15.replace_io("t_motorstart", "?", bit=5)
        self.rpi.io.pbit8_15.replace_io("t_magazin1", "?", bit=6)
        self.rpi.io.pbit8_15.replace_io("t_magazin2", "?", bit=7)

        # Panel Eingang Byte 7 + 8
        self.rpi.io.p_drehzahl1.replace_io("p_drehzahl", "H")

        # Panel Ausgänge Byte 0
        self.rpi.io.meldung0_7.replace_io("err_master", "?", bit=0)
        self.rpi.io.meldung0_7.replace_io("err_funichtaus", "?", bit=1)
        self.rpi.io.meldung0_7.replace_io("err_magazin1", "?", bit=2)
        self.rpi.io.meldung0_7.replace_io("err_magazin2", "?", bit=3)
        self.rpi.io.meldung0_7.replace_io("err_rutsche", "?", bit=4)
        self.rpi.io.meldung0_7.replace_io("err_endlage", "?", bit=5)

        # Panel Ausgänge Byte 1
        self.rpi.io.meldung8_15.replace_io("warn_mag1full", "?", bit=0)
        self.rpi.io.meldung8_15.replace_io("warn_mag2full", "?", bit=1)
        self.rpi.io.meldung8_15.replace_io("aktivauto", "?", bit=6)
        self.rpi.io.meldung8_15.replace_io("aktivhand", "?", bit=7)

        # Globale Steuerungsthreads vorbereiten
        funccatalog.anlage = funccatalog.Anlage(self.rpi)
        funccatalog.band = funccatalog.Band(self.rpi)
        funccatalog.zylinder = funccatalog.Zylinder(self.rpi)

        # Funktions Thrads vorbereiten
        self.th_automatik = AutomatikPara(self.rpi)
        self.th_hand = Hand(self.rpi)

        # Event für Start- und Stopptaster registrieren
        self.rpi.io.t_start.reg_event(
            self.evt_start, edge=revpimodio2.RISING
        )
        self.rpi.io.t_stop.reg_event(
            self.evt_stop, edge=revpimodio2.FALLING
        )

        # Fehlermeldungen registrieren
        self.rpi.io.notaus_ok.reg_event(
            self.evt_fehler, edge=revpimodio2.FALLING
        )
        self.rpi.io.motorschutz_ok.reg_event(
            self.evt_fehler, edge=revpimodio2.FALLING
        )
        self.rpi.io.fu_ok.reg_event(
            self.evt_fehler, edge=revpimodio2.FALLING
        )
        self.rpi.io.m_druck_ok.reg_event(
            self.evt_fehler, delay=2000, edge=revpimodio2.FALLING
        )

    def evt_fehler(self, ioname, iovalue):
        """Eventfunktion fuer Druckabfall.
        @param ioname Name des IO, der Event ausloest
        @param iovalue Wert des IO, der Event ausloest"""
        self.rpi.io.err_master.value = True
        if ioname == "notaus_ok":
            self.th_hand.stop()
        self.th_automatik.stop()
        funccatalog.anlage.set_stoerung(ioname)

    def evt_start(self, ioname, iovalue):
        """Eventfunktion fuer den Taster t_start.

        @param ioname Name des IO, der Event ausloest
        @param iovalue Wert des IO, der Event ausloest
        @return None

        """
        # Prüfung auf Notaus
        if not self.rpi.io.notaus_ok.value:
            return None

        # Prüfen ob Threads schon laufen
        if not (self.th_automatik.is_alive() or self.th_hand.is_alive()):

            if self.rpi.io.t_automatik.value:
                # Automatikbetrieb starten
                if not self.rpi.io.err_master.value:
                    self.th_automatik = AutomatikPara(self.rpi)
                    self.th_automatik.start()

            else:
                # Handbetrieb starten
                self.th_hand = Hand(self.rpi)
                self.th_hand.start()

    def evt_stop(self, ioname, iovalue):
        """Beendet das Revolutionsumbauprogramm."""
        # Laufende Threads beenden
        self.th_automatik.stop()
        self.th_hand.stop()

    def start(self):
        """Startet das Revolutionsumbauprogramm."""
        # LED A1 grün schalten
        self.rpi.core.A1 = 1

        # Druckventil öffnen
        self.rpi.io.v_druck.value = True
        self.rpi.io.m_druck_ok.wait(okvalue=True, timeout=3000)

        # Alle Funktionsthrads starten
        funccatalog.anlage.start()
        funccatalog.band.start()
        funccatalog.zylinder.start()

        # In den mainloop() gehen und auf Events warten
        self.rpi.mainloop()

        # LED A1 ausschalten
        self.rpi.core.A1 = 0

        # Druckventil schließen
        self.rpi.io.v_druck.value = False

        # Prozessabbild schreiben, da autorefresh hier nicht mehr läuft
        self.rpi.writeprocimg()

        # RevPiModIO herunterfahren
        self.rpi.cleanup()

        print("alles beendet")

    def stop(self, signum=None, frame=None):
        """Beendet das Revolutionsumbauprogramm."""
        # Alle Steuerungsthreads beenden
        self.th_automatik.stop()
        self.th_hand.stop()

        # Alle Funktionsthrads beenden
        funccatalog.anlage.stop()
        funccatalog.band.stop()
        funccatalog.zylinder.stop()

        # mainloop() verlassen und autorefresh beenden
        self.rpi.exit()


if __name__ == "__main__":
    root = RevPiRevolution()
    root.start()
